# Exercise 9.1 "Finite Element Analysis of Composite Materials : with Abaqus" 
# using ugens-std.obj
# load the Python Modules
# updated for Abaqus 2017 (R17)
from abaqusConstants import *
from mesh import *
from step import *
from regionToolset import Region
from multiprocessing import cpu_count
from visualization import openOdb
from abaqus import mdb
import csv                  # utilities to write a .CSV file
from UgenKeyword import *   # utilities to write the UGENS parameters on the Job.inp file directly from CAE
session.journalOptions.setValues(replayGeometry=COORDINATE, recoverGeometry=COORDINATE)

# Enter your model parameters here 
NL = 3                                  # number of laminas (scalar)
ModelParameters = {}                    # define a dictionary data tye
ModelParameters['E1']       = '44700'   # add to the dictionary
ModelParameters['E2']       = '12700'
ModelParameters['G12']      =  '5800'
ModelParameters['Nu12']     = '.297'
ModelParameters['Nu23']     = '.41'
ModelParameters['Cte1']     = '3.7'
ModelParameters['Cte2']     = '30.'
ModelParameters['GIIc']     = '1e16'
ModelParameters['GIc']      = '0.254'
ModelParameters['tk']       = '.144'
ModelParameters['DeltaT']   = '0'
ModelParameters['Lss']      = [(1,0), (8,90), (.5,0)] # a [list] of tuples (thickness angle)
ModelParameters['Strain'] = '1.8735'    # max. strain to run the model in %
ShellDimensionX = 10.                   # model dimensions
ShellDimensionY = 10.                   # model dimensions

# Generate the Abaqus model    
mdb.close()
Model = mdb.models['Model-1']

# shell dimensions
Sketch = Model.ConstrainedSketch(name='LaminateSketch', sheetSize=10.*ShellDimensionX)
Sketch.rectangle(point1 = (0.0, 0.0),
                 point2 = (ShellDimensionX, ShellDimensionX)
                 )

Part = Model.Part(dimensionality = THREE_D,
                  name           = 'Laminate',
                  type           = DEFORMABLE_BODY
                  )
                  
Part.BaseShell(sketch=Sketch)

Instance = Model.rootAssembly.Instance(dependent = ON,
                                       name='Laminate-1', 
                                       part=Part
                                       )

Part.Set(faces=Part.faces,
         name='WholeLaminate'
         )

# STEP
# minInc, reduce it as much as needed
# maxNumInc, set to a bigger number to allow for reduced minInc
Step = Model.StaticStep(timePeriod=float(ModelParameters['Strain']),
                        initialInc=0.01,
                        minInc=0.0001,
                        maxInc=0.01,
                        maxNumInc=1000,
                        name='ApplyStrain',
                        previous='Initial'
                        )
                        
Model.fieldOutputRequests['F-Output-1'].setValues(variables=('U', 'RF', 'SDV'))
                        
Model.XsymmBC(createStepName='Initial',
              name='SymmetryX', 
              region=Region(edges=Instance.edges.findAt(((0.0, ShellDimensionY/2., 0.0), ), ))
              )

Model.YsymmBC(createStepName='Initial',
              name='SymmetryY', 
              region=Region(edges=Instance.edges.findAt(((ShellDimensionX/2., 0.0, 0.0), ), ))
              )

Model.EncastreBC(createStepName='Initial',
                 name='NoRigidBodyMotion',
                 region=Region(vertices=Instance.vertices.findAt(((0.0, 0.0, 0.0), ), ))
                 )
                 
Model.DisplacementBC(createStepName='ApplyStrain',
                     name='Strain',
                     region=Region(edges=Instance.edges.findAt(((ShellDimensionX, ShellDimensionY/2., 0.0), ), )),
                     u1=ShellDimensionX*float(ModelParameters['Strain'])/100.
                     )

# membrane only
mdb.models['Model-1'].DisplacementBC(amplitude=UNSET, createStepName='Initial', 
    distributionType=UNIFORM, fieldName='', localCsys=None, name='BC-4', 
    region=Region(
    faces=mdb.models['Model-1'].rootAssembly.instances['Laminate-1'].faces.findAt(
    ((ShellDimensionX/2., ShellDimensionY/2., 0.0), (0.0, 0.0, 0.0)), )), u1=UNSET, u2=UNSET, u3=
    SET, ur1=SET, ur2=SET, ur3=SET)

# CONTROLS
# Abaqus default R_n^alpha=0.005 (first value in the list); changed to 0.02 
# R_n^alpha: convergence criterion for ratio of largest residual to corresponding average flux norm
mdb.models['Model-1'].steps['ApplyStrain'].control.setValues(
    allowPropagation=OFF, resetDefaultValues=OFF, displacementField=(0.02, 
    0.01, 0.0, 0.0, 0.02, 1e-05, 0.001, 1e-08, 1.0, 1e-05, 1e-08), 
    electricalPotentialField=(0.005, 0.01, 0.0, 0.0, 0.02, 1e-05, 0.001, 1e-08, 
    1.0, 1e-05), hydrostaticFluidPressureField=(0.005, 0.01, 0.0, 0.0, 0.02, 
    1e-05, 0.001, 1e-08, 1.0, 1e-05), rotationField=(0.005, 0.01, 0.0, 0.0, 
    0.02, 1e-05, 0.001, 1e-08, 1.0, 1e-05))

Part.setElementType(elemTypes=(ElemType(elemCode=S4R), ), 
                    regions=Region(faces = Part.faces)
                    )

Part.seedPart(size=ShellDimensionX) # only one element

Part.generateMesh()

# Abaqus does not allow UGENS with multiple Cpus
# Do not use numCpus=cpu_count(). Use numCpus=1
Job = mdb.Job(name='Job', 
              model='Model-1',
              nodalOutputPrecision=FULL,
              numCpus=1,
              numDomains=cpu_count(),
              userSubroutine='ugens-std.obj'
              )

Model.keywordBlock.synchVersions(storeNodesAndElements=False)
Model.keywordBlock.insert(5, UgenKeyword(ModelParameters))
Model.keywordBlock.insert(30, """
*Initial Conditions, type=SOLUTION, USER
""")

mdb.jobs['Job'].writeInput(consistencyChecking=OFF)

# execution starts here
Job.submit()
Job.waitForCompletion()

# post-processing starts here
# be extremely careful with the blank lines and indentation required by Python (thanks Python!)
Results = openOdb(path='Job.odb')  
NL = len(ModelParameters['Lss'])
WholeLaminateThickness = 2.*sum([x[0] for x in ModelParameters['Lss']])*float(ModelParameters['tk'])
ofile = open('Job.csv', 'wb')
Writer = csv.writer(ofile)
Header = ['Strain in X [%]', 'Stress in X [MPa]', 'Ex [MPa]', 'PRxy']
for i in range(NL):
    Header.append('Crack Density '+str(i+1)+' [1/mm]')

Writer.writerow(Header)
for fn in range(1, len(Results.steps['ApplyStrain'].frames)):
    frame  = Results.steps['ApplyStrain'].frames[fn]
    StrainX =     frame.fieldOutputs['U' ].values[3].dataDouble[0]/ShellDimensionX
    StrainY =     frame.fieldOutputs['U' ].values[3].dataDouble[1]/ShellDimensionX
    StressX = 2.0*frame.fieldOutputs['RF'].values[3].dataDouble[0]/(ShellDimensionX*WholeLaminateThickness)
    
    Row = [StrainX*100., StressX, StressX/StrainX, -StrainY/StrainX]
    for i in range(NL):
        Row.append(frame.fieldOutputs['SDV'+str(3*i+1)].values[0].dataDouble)#updated for R17
    
    Writer.writerow(Row)

ofile.close()
Results.close()
mdb.close()
